import * as bigquery from "@google-cloud/bigquery";
import { FirestoreEventHistoryTracker, FirestoreDocumentChangeEvent } from "../tracker";
export { RawChangelogSchema, RawChangelogViewSchema } from "./schema";
export interface FirestoreBigQueryEventHistoryTrackerConfig {
    datasetId: string;
    tableId: string;
    datasetLocation?: string | undefined;
    transformFunction?: string | undefined;
    timePartitioning?: string | undefined;
    timePartitioningField?: string | undefined;
    timePartitioningFieldType?: string | undefined;
    timePartitioningFirestoreField?: string | undefined;
    clustering: string[] | null;
    databaseId?: string | undefined;
    wildcardIds?: boolean;
    bqProjectId?: string | undefined;
    backupTableId?: string | undefined;
    useNewSnapshotQuerySyntax?: boolean;
    skipInit?: boolean;
    kmsKeyName?: string | undefined;
    useMaterializedView?: boolean;
    useIncrementalMaterializedView?: boolean;
    maxStaleness?: string;
    refreshIntervalMinutes?: number;
}
/**
 * An FirestoreEventHistoryTracker that exports data to BigQuery.
 *
 * When the first event is received, it creates necessary BigQuery resources:
 * - Dataset: {@link FirestoreBigQueryEventHistoryTrackerConfig#datasetId}.
 * - Table: Raw change log table {@link FirestoreBigQueryEventHistoryTracker#rawChangeLogTableName}.
 * - View: Latest view {@link FirestoreBigQueryEventHistoryTracker#rawLatestView}.
 * If any subsequent data export fails, it will attempt to reinitialize.
 */
export declare class FirestoreBigQueryEventHistoryTracker implements FirestoreEventHistoryTracker {
    config: FirestoreBigQueryEventHistoryTrackerConfig;
    bq: bigquery.BigQuery;
    _initialized: boolean;
    constructor(config: FirestoreBigQueryEventHistoryTrackerConfig);
    record(events: FirestoreDocumentChangeEvent[]): Promise<void>;
    private transformRows;
    serializeData(eventData: any): any;
    /**
     * Check whether a failed operation is retryable or not.
     * Reasons for retrying:
     * 1) We added a new column to our schema. Sometimes BQ is not ready to stream insertion records immediately
     *    after adding a new column to an existing table (https://issuetracker.google.com/35905247)
     */
    private isRetryableInsertionError;
    /**
     * Tables can often take time to create and propagate.
     * A half a second delay is added per check while the function
     * continually re-checks until the referenced dataset and table become available.
     */
    private _waitForInitialization;
    /**
     * Inserts rows of data into the BigQuery raw change log table.
     */
    private insertData;
    /**
     * Creates the BigQuery resources with the expected schema for {@link FirestoreEventHistoryTracker}.
     * After the first invokation, it skips initialization assuming these resources are still there.
     */
    initialize(): Promise<void>;
    /**
     * Creates the specified dataset if it doesn't already exists.
     */
    private initializeDataset;
    /**
     * Creates the raw change log table if it doesn't already exist.
     */
    private initializeRawChangeLogTable;
    /**
     * Creates the latest snapshot view, which returns only latest operations
     * of all existing documents over the raw change log table.
     */
    private _initializeLatestView;
    bigqueryDataset(): bigquery.Dataset;
    private rawChangeLogTableName;
    private rawLatestView;
}
