"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FirestoreBigQueryEventHistoryTracker = exports.RawChangelogViewSchema = exports.RawChangelogSchema = void 0;
const bigquery = require("@google-cloud/bigquery");
const firestore_1 = require("firebase-admin/firestore");
const traverse = require("traverse");
const node_fetch_1 = require("node-fetch");
const schema_1 = require("./schema");
const handleFailedTransactions_1 = require("./handleFailedTransactions");
const tracker_1 = require("../tracker");
const logs = require("../logs");
const partitioning_1 = require("./partitioning");
const clustering_1 = require("./clustering");
const checkUpdates_1 = require("./checkUpdates");
const utils_1 = require("./utils");
const initializeLatestView_1 = require("./initializeLatestView");
var schema_2 = require("./schema");
Object.defineProperty(exports, "RawChangelogSchema", { enumerable: true, get: function () { return schema_2.RawChangelogSchema; } });
Object.defineProperty(exports, "RawChangelogViewSchema", { enumerable: true, get: function () { return schema_2.RawChangelogViewSchema; } });
/**
 * An FirestoreEventHistoryTracker that exports data to BigQuery.
 *
 * When the first event is received, it creates necessary BigQuery resources:
 * - Dataset: {@link FirestoreBigQueryEventHistoryTrackerConfig#datasetId}.
 * - Table: Raw change log table {@link FirestoreBigQueryEventHistoryTracker#rawChangeLogTableName}.
 * - View: Latest view {@link FirestoreBigQueryEventHistoryTracker#rawLatestView}.
 * If any subsequent data export fails, it will attempt to reinitialize.
 */
class FirestoreBigQueryEventHistoryTracker {
    constructor(config) {
        this.config = config;
        this._initialized = false;
        this.bq = new bigquery.BigQuery();
        this.bq.projectId = config.bqProjectId || process.env.PROJECT_ID;
        if (!this.config.datasetLocation) {
            this.config.datasetLocation = "us";
        }
    }
    async record(events) {
        if (!this.config.skipInit) {
            await this.initialize();
        }
        const partitionHandler = new partitioning_1.Partitioning(this.config);
        const rows = events.map((event) => {
            const partitionValue = partitionHandler.getPartitionValue(event);
            const { documentId, ...pathParams } = event.pathParams || {};
            return {
                insertId: event.eventId,
                json: {
                    timestamp: event.timestamp,
                    event_id: event.eventId,
                    document_name: event.documentName,
                    document_id: event.documentId,
                    operation: tracker_1.ChangeType[event.operation],
                    data: JSON.stringify(this.serializeData(event.data)),
                    old_data: event.oldData
                        ? JSON.stringify(this.serializeData(event.oldData))
                        : null,
                    ...partitionValue,
                    ...(this.config.wildcardIds &&
                        event.pathParams && { path_params: JSON.stringify(pathParams) }),
                },
            };
        });
        const transformedRows = await this.transformRows(rows);
        await this.insertData(transformedRows);
    }
    async transformRows(rows) {
        if (this.config.transformFunction && this.config.transformFunction !== "") {
            const response = await (0, node_fetch_1.default)(this.config.transformFunction, {
                method: "post",
                body: JSON.stringify({ data: rows }),
                headers: { "Content-Type": "application/json" },
            });
            const responseJson = await response.json();
            // To support callable functions, first check result.data
            return responseJson?.result?.data ?? responseJson.data;
        }
        return rows;
    }
    serializeData(eventData) {
        if (typeof eventData === "undefined") {
            return undefined;
        }
        const data = traverse(eventData).map(function (property) {
            if (property && property.constructor) {
                if (property.constructor.name === "Buffer") {
                    this.remove();
                }
                if (property.constructor.name === firestore_1.DocumentReference.name) {
                    this.update(property.path);
                }
            }
        });
        return data;
    }
    /**
     * Check whether a failed operation is retryable or not.
     * Reasons for retrying:
     * 1) We added a new column to our schema. Sometimes BQ is not ready to stream insertion records immediately
     *    after adding a new column to an existing table (https://issuetracker.google.com/35905247)
     */
    async isRetryableInsertionError(e) {
        let isRetryable = true;
        const expectedErrors = [
            { message: "no such field.", location: schema_1.documentIdField.name },
            { message: "no such field.", location: schema_1.documentPathParams.name },
        ];
        if (e.response &&
            e.response.insertErrors &&
            e.response.insertErrors.errors) {
            const errors = e.response.insertErrors.errors;
            errors?.forEach((error) => {
                let isExpected = false;
                expectedErrors?.forEach((expectedError) => {
                    if (error.message === expectedError.message &&
                        error.location === expectedError.location) {
                        isExpected = true;
                    }
                });
                if (!isExpected) {
                    isRetryable = false;
                }
            });
        }
        return isRetryable;
    }
    /**
     * Tables can often take time to create and propagate.
     * A half a second delay is added per check while the function
     * continually re-checks until the referenced dataset and table become available.
     */
    async _waitForInitialization() {
        const dataset = this.bigqueryDataset();
        const changelogName = this.rawChangeLogTableName();
        let materializedViewName;
        if (this.config.useMaterializedView) {
            materializedViewName = this.rawLatestView();
        }
        return (0, utils_1.waitForInitialization)({
            dataset,
            changelogName,
            materializedViewName,
        });
    }
    /**
     * Inserts rows of data into the BigQuery raw change log table.
     */
    async insertData(rows, overrideOptions = {}, retry = true) {
        const options = {
            skipInvalidRows: false,
            ignoreUnknownValues: false,
            raw: true,
            ...overrideOptions,
        };
        try {
            const dataset = this.bigqueryDataset();
            const table = dataset.table(this.rawChangeLogTableName());
            logs.dataInserting(rows.length);
            await table.insert(rows, options);
            logs.dataInserted(rows.length);
        }
        catch (e) {
            if (retry && this.isRetryableInsertionError(e)) {
                retry = false;
                logs.dataInsertRetried(rows.length);
                return this.insertData(rows, { ...overrideOptions, ignoreUnknownValues: true }, retry);
            }
            // Exceeded number of retries, save in failed collection
            if (!retry && this.config.backupTableId) {
                await (0, handleFailedTransactions_1.default)(rows, this.config, e);
            }
            // Reinitializing in case the destintation table is modified.
            this._initialized = false;
            logs.bigQueryTableInsertErrors(e.errors);
            throw e;
        }
    }
    /**
     * Creates the BigQuery resources with the expected schema for {@link FirestoreEventHistoryTracker}.
     * After the first invokation, it skips initialization assuming these resources are still there.
     */
    async initialize() {
        try {
            if (this._initialized) {
                return;
            }
            try {
                await this.initializeDataset();
            }
            catch (error) {
                const message = (0, utils_1.parseErrorMessage)(error, "initializing dataset");
                throw new Error(`Error initializing dataset: ${message}`);
            }
            try {
                await this.initializeRawChangeLogTable();
            }
            catch (error) {
                const message = (0, utils_1.parseErrorMessage)(error, "initializing raw change log table");
                throw new Error(`Error initializing raw change log table: ${message}`);
            }
            try {
                await this._initializeLatestView();
            }
            catch (error) {
                const message = (0, utils_1.parseErrorMessage)(error, "initializing latest view");
                throw new Error(`Error initializing latest view: ${message}`);
            }
            await this._waitForInitialization();
            this._initialized = true;
        }
        catch (error) {
            const message = (0, utils_1.parseErrorMessage)(error, "initializing BigQuery resources");
            console.error("Error initializing BigQuery resources: ", message);
            throw error;
        }
    }
    /**
     * Creates the specified dataset if it doesn't already exists.
     */
    async initializeDataset() {
        const dataset = this.bigqueryDataset();
        const [datasetExists] = await dataset.exists();
        if (datasetExists) {
            logs.bigQueryDatasetExists(this.config.datasetId);
        }
        else {
            try {
                logs.bigQueryDatasetCreating(this.config.datasetId);
                await dataset.create();
                logs.bigQueryDatasetCreated(this.config.datasetId);
            }
            catch (ex) {
                logs.tableCreationError(this.config.datasetId, ex.message);
            }
        }
        return dataset;
    }
    /**
     * Creates the raw change log table if it doesn't already exist.
     */
    async initializeRawChangeLogTable() {
        const changelogName = this.rawChangeLogTableName();
        const dataset = this.bigqueryDataset();
        const table = dataset.table(changelogName);
        const [tableExists] = await table.exists();
        const partitioning = new partitioning_1.Partitioning(this.config, table);
        const clustering = new clustering_1.Clustering(this.config, table);
        if (tableExists) {
            logs.bigQueryTableAlreadyExists(table.id, dataset.id);
            const [metadata] = await table.getMetadata();
            const fields = metadata.schema ? metadata.schema.fields : [];
            await clustering.updateClustering(metadata);
            const documentIdColExists = fields.find((column) => column.name === "document_id");
            const pathParamsColExists = fields.find((column) => column.name === "path_params");
            const oldDataColExists = fields.find((column) => column.name === "old_data");
            if (!oldDataColExists) {
                fields.push(schema_1.oldDataField);
                logs.addNewColumn(this.rawChangeLogTableName(), schema_1.oldDataField.name);
            }
            if (!documentIdColExists) {
                fields.push(schema_1.documentIdField);
                logs.addNewColumn(this.rawChangeLogTableName(), schema_1.documentIdField.name);
            }
            if (!pathParamsColExists && this.config.wildcardIds) {
                fields.push(schema_1.documentPathParams);
                logs.addNewColumn(this.rawChangeLogTableName(), schema_1.documentPathParams.name);
            }
            /** Updated table metadata if required */
            const shouldUpdate = await (0, checkUpdates_1.tableRequiresUpdate)({
                table,
                config: this.config,
                documentIdColExists,
                pathParamsColExists,
                oldDataColExists,
            });
            if (shouldUpdate) {
                /** set partitioning */
                await partitioning.addPartitioningToSchema(metadata.schema.fields);
                /** update table metadata with changes. */
                await table.setMetadata(metadata);
                logs.updatingMetadata(this.rawChangeLogTableName(), {
                    config: this.config,
                    documentIdColExists,
                    pathParamsColExists,
                    oldDataColExists,
                });
            }
        }
        else {
            logs.bigQueryTableCreating(changelogName);
            const schema = { fields: [...schema_1.RawChangelogSchema.fields] };
            if (this.config.wildcardIds) {
                schema.fields.push(schema_1.documentPathParams);
            }
            const options = { friendlyName: changelogName, schema };
            if (this.config.kmsKeyName) {
                options["encryptionConfiguration"] = {
                    kmsKeyName: this.config.kmsKeyName,
                };
            }
            //Add partitioning
            await partitioning.addPartitioningToSchema(schema.fields);
            await partitioning.updateTableMetadata(options);
            // Add clustering
            await clustering.updateClustering(options);
            try {
                await table.create(options);
                logs.bigQueryTableCreated(changelogName);
            }
            catch (ex) {
                logs.tableCreationError(changelogName, ex.message);
            }
        }
        return table;
    }
    /**
     * Creates the latest snapshot view, which returns only latest operations
     * of all existing documents over the raw change log table.
     */
    async _initializeLatestView() {
        const dataset = this.bigqueryDataset();
        const view = dataset.table(this.rawLatestView());
        const [viewExists] = await view.exists();
        return await (0, initializeLatestView_1.initializeLatestView)({
            bq: this.bq,
            changeTrackerConfig: this.config,
            dataset,
            view,
            viewExists,
            rawChangeLogTableName: this.rawChangeLogTableName(),
            rawLatestViewName: this.rawLatestView(),
        });
    }
    bigqueryDataset() {
        return this.bq.dataset(this.config.datasetId, {
            location: this.config.datasetLocation,
        });
    }
    rawChangeLogTableName() {
        return `${this.config.tableId}_raw_changelog`;
    }
    rawLatestView() {
        return `${this.config.tableId}_raw_latest`;
    }
}
exports.FirestoreBigQueryEventHistoryTracker = FirestoreBigQueryEventHistoryTracker;
