"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeLatestMaterializedView = exports.shouldRecreateMaterializedView = void 0;
const logs = require("../logs");
const snapshot_1 = require("./snapshot");
const firebase_functions_1 = require("firebase-functions");
const sqlFormatter = require("sql-formatter");
async function shouldRecreateMaterializedView(view, config, source) {
    const [viewMetadata] = await view.getMetadata();
    const isIncremental = !viewMetadata.materializedView
        ?.allowNonIncrementalDefinition;
    const incrementalMatch = isIncremental === !!config.useIncrementalMaterializedView;
    const viewQuery = viewMetadata.materializedView?.query || "";
    const queryMatch = sqlFormatter.format(viewQuery) === sqlFormatter.format(source);
    return !queryMatch || !incrementalMatch;
}
exports.shouldRecreateMaterializedView = shouldRecreateMaterializedView;
/**
 * Creates the latest materialized view.
 */
async function initializeLatestMaterializedView({ bq, changeTrackerConfig: config, view, viewExists, rawChangeLogTableName, rawLatestViewName, schema, }) {
    try {
        const { query, source } = config.useIncrementalMaterializedView
            ? (0, snapshot_1.buildMaterializedViewQuery)({
                projectId: bq.projectId,
                datasetId: config.datasetId,
                tableName: rawChangeLogTableName,
                rawLatestViewName,
                schema,
            })
            : (0, snapshot_1.buildNonIncrementalMaterializedViewQuery)({
                projectId: bq.projectId,
                datasetId: config.datasetId,
                tableName: rawChangeLogTableName,
                maxStaleness: config.maxStaleness,
                refreshIntervalMinutes: config.refreshIntervalMinutes,
                rawLatestViewName,
                enableRefresh: true,
                schema,
            });
        const desiredQuery = sqlFormatter.format(query);
        if (viewExists) {
            const shouldRecreate = await shouldRecreateMaterializedView(view, config, source);
            if (!shouldRecreate) {
                firebase_functions_1.logger.warn(`Materialized view requested, but a view with matching configuration exists. Skipping creation.`);
                return view;
            }
            firebase_functions_1.logger.warn(`Configuration mismatch detected for ${rawLatestViewName} ` +
                `Recreating view...`);
            await view.delete();
            return await initializeLatestMaterializedView({
                bq,
                changeTrackerConfig: config,
                view,
                viewExists: false,
                rawChangeLogTableName,
                rawLatestViewName,
                schema,
            });
        }
        logs.bigQueryViewCreating(rawLatestViewName, desiredQuery);
        await bq.query(desiredQuery);
        logs.bigQueryViewCreated(rawLatestViewName);
    }
    catch (error) {
        logs.tableCreationError(rawLatestViewName, error.message);
        throw error;
    }
    return view;
}
exports.initializeLatestMaterializedView = initializeLatestMaterializedView;
