"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeLatestView = void 0;
const schema_1 = require("./schema");
const logs = require("../logs");
const snapshot_1 = require("./snapshot");
const checkUpdates_1 = require("./checkUpdates");
const initializeLatestMaterializedView_1 = require("./initializeLatestMaterializedView");
/**
 * Creates the latest snapshot view or materialized view.
 */
async function initializeLatestView({ changeTrackerConfig: config, dataset, view, viewExists, rawChangeLogTableName, rawLatestViewName, bq, }) {
    if (config.useMaterializedView) {
        const schema = { fields: [...schema_1.RawChangelogViewSchema.fields] };
        if (config.wildcardIds) {
            schema.fields.push(schema_1.documentPathParams);
        }
        return (0, initializeLatestMaterializedView_1.initializeLatestMaterializedView)({
            bq,
            changeTrackerConfig: config,
            view,
            viewExists,
            rawChangeLogTableName,
            rawLatestViewName,
            schema,
        });
    }
    const schema = schema_1.RawChangelogViewSchema;
    if (viewExists) {
        logs.bigQueryViewAlreadyExists(view.id, dataset.id);
        const [metadata] = await view.getMetadata();
        const fields = (metadata.schema ? metadata.schema.fields : []);
        if (config.wildcardIds) {
            schema.fields.push(schema_1.documentPathParams);
        }
        const columnNames = fields.map((field) => field.name);
        const documentIdColExists = columnNames.includes("document_id");
        const pathParamsColExists = columnNames.includes("path_params");
        const oldDataColExists = columnNames.includes("old_data");
        const updateView = (0, checkUpdates_1.viewRequiresUpdate)({
            metadata,
            config,
            documentIdColExists,
            pathParamsColExists,
            oldDataColExists,
        });
        if (updateView) {
            metadata.view = (0, snapshot_1.latestConsistentSnapshotView)({
                datasetId: config.datasetId,
                tableName: rawChangeLogTableName,
                schema,
                useLegacyQuery: !config.useNewSnapshotQuerySyntax,
            });
            if (!documentIdColExists) {
                logs.addNewColumn(rawLatestViewName, schema_1.documentIdField.name);
            }
            await view.setMetadata(metadata);
            logs.updatingMetadata(rawLatestViewName, {
                config,
                documentIdColExists,
                pathParamsColExists,
                oldDataColExists,
            });
        }
    }
    else {
        const schema = { fields: [...schema_1.RawChangelogViewSchema.fields] };
        if (config.wildcardIds) {
            schema.fields.push(schema_1.documentPathParams);
        }
        const latestSnapshot = (0, snapshot_1.latestConsistentSnapshotView)({
            datasetId: config.datasetId,
            tableName: rawChangeLogTableName,
            schema,
            bqProjectId: bq.projectId,
            useLegacyQuery: !config.useNewSnapshotQuerySyntax,
        });
        logs.bigQueryViewCreating(rawLatestViewName, latestSnapshot.query);
        const options = {
            friendlyName: rawLatestViewName,
            view: latestSnapshot,
        };
        try {
            await view.create(options);
            await view.setMetadata({ schema: schema_1.RawChangelogViewSchema });
            logs.bigQueryViewCreated(rawLatestViewName);
        }
        catch (error) {
            logs.tableCreationError(rawLatestViewName, error.message);
        }
    }
    return view;
}
exports.initializeLatestView = initializeLatestView;
